#!/usr/bin/env python
#
# pyirc : An IRC pure-module for Python
# 
# Copyright (c) 2003 Sylvain Hellegouarch.  Distributed under GNU GPL.
#
# Authors: Sylvain Hellegouarch (sh@programmationworld.com)
#

#
# Imported modules
# All these modules are provided within the core distribution of Python 2.2.x
#
import socket
import string
import re
import time
import os.path
from types import *
from Queue import *
from struct import pack, unpack

#
# Some global data
#
_message_sep = re.compile("\r\n")
_rfc_1459_command_regexp = re.compile("^(:(?P<prefix>[^ ]+) +)?(?P<command>[^ ]+)( *(?P<argument> .+))?")
_error_range = []
for val in range(401,503,1):
    _error_range.append(str(val))
_x01 = chr(1)
DCC_PACKET_SIZE = 1024

#
# Ths dictionnary is just a memo for the different messages that could be put into the queue and their format
# prefix  : means either the server name or the username with following format ~username@host (~ is not compulsory)
# arguments : data linked to the message (might need to be parsed again sometimes)
#

_managed_messages_received = {
    'privmsg'      : "privmsg,server_name,prefix,from_nickname,from_channel,arguments",
    'ctcp'        : "ctcp, server_name, prefix,from_nickname,to_nickname,ctcp_command,arguments",
    'dccsend'     : "dccsend,server_name,prefix,from_nickname,file_name,from_ip,on_port,file_size",
    'dccchat'     : "dccchat,server_name,prefix,from_nickname,from_ip,on_port",
    'dccchatok'   : "dccchatok,server_name,with_user,socket_object",
    'read'        : "read,nickname,data", #from a DCC chat connection
    'closed'      : "closed,nickname", #from a DCC chat connection
    'notice'      : "privmsg,server_name,prefix,from_nickname,from_channel,arguments",
    'join'        : "join,server_name,prefix,from_nickname,arguments",
    'part'        : "part,server_name,prefix,from_nickname,from_channel,arguments",
    'quit'        : "quit,server_name,prefix,from_nickname,arguments",
    'kick'        : "kick,server_name,prefix,from_nickname,from_channel,arguments",
    'ping'        : "You've received a ping request",
    'topic'       : "topic,server_name,prefix,from_nickname,from_channel,arguments",
    'notopic'     : "notopic,server_name,prefix,from_nickname,from_channel,arguments",
    'nick'        : "nick,server_name,prefix,old_nick,new_nick",
    'datanames'   : "datanames,server_name,from_channel,nicks",
    'endnames'    : "endnames,server_name,arguments",
    'startmotd'   : "startmotd,server_name,arguments",
    'datamotd'    : "datamotd,server_name,arguments",
    'endmotd'     : "endmotd,server_name,arguments",    
    'init_info'   : "init_info,server_name,command_id,arguments",
    'luser'       : "luser,server_name,command_id,arguments",
    'datalinks'   : "datalinks,server_name,mask,server,hop,server_desc",
    'endlinks'    : "endlinks,server_name,server,arguments",
    'ison'        : "ison,server_name,nicks",
    'mode'        : "mode,server_name,prefix,from_nickname,mode,nick",
    'startlist'   : "starts_list,server_name,arguments",
    'datalist'    : "datalist,server_name,channel,nb_of_visisble_users,topic",
    'endlist'     : "endlist,server_name,arguments",
    'time'        : "time,server_name,prefix,time_of_the_server",
    'version'     : "version,server_name,from_nickname,version_of_the_server(.debuglevel),server,comments",
    'invite'      : "invite,server_name,prefix,from_nickname,to_nickname,to_channel",
    'datainfo'    : "datainfo,server_name,from_nickname,arguments",
    'endinfo'     : "endinfo,server_name,from_nickname,arguments",
    'userhost'    : "userhost,server_name,from_nickname,arguments",
    'datawho'     : "datawho,server_name,channel,username,host,H|G[*][@|+],hop,real_name",
    'endwho'      : "endwho,server_name,nickname,arguments",
    'datawhois'   : "datawhois,server_name,command,nickname,arguments",
    'endwhois'    : "endwhois,server_name,nickname,arguments",
    'datawhowas'  : "datawhowas,server_name,nickname,username,host,real_name",
    'endwhowas'   : "endwhowas,server_name,arguments",
    'admin'       : "admin,server_name,from_nickname,arguments",
    'datatrace'   : "datatrace,server_name,command_id,arguments",
    'endtrace'    : "endtrace,server_name,data,arguments",
    'markedaway'  : "markedaway,server_name,from_nickname,arguments",
    'unmarkedaway': "unmarkedaway,server_name,from_nickname,arguments",
    'isaway'      : "isaway,server_name,from_nickname,who_is_away,message_sent",
    'tryagain'    : "",
    'error'       : "error,server_name,prefix,command,request,arguments",
    'unknown'     : "unknown,server_name,prefix,from_nickname,from_channel,argeuments"
}

#
# A base class for IRC errors
# Does nothing and is never actually returned
#
class IRCError(Exception):
    """An IRC error"""
    pass

#
# Connection error class
#
class IRCConnectionError(IRCError):
    def __init__(self, value):
        """Define a connection error to a server"""
        self.value = value
    def __str__(self):
        return repr(self.value)

#
# DCCChat is the class which wraps the DCC chat "protocol". A DCC chat is on when two users are directly connected over a TCP/IP socket connection.
# This class is easy but has two different way to be used.
#
# First if you are the user who requests the chat, you'll just create a DCCCChat object passing only the nick of the target user. Then you must call the wait_for_connexion() method but not the connect() one, it would be pointless. The best way would be to call it in a separate thread as it will create a socket and wait to accept a connexion on it which will block the current processus. Once the connection has been accepted, the method will put a 'dccchatok' message on the queue of the server so that it will know that it can proceed with the chat connection further (you'll need to call get_sock_id() to retrieve the socket id).
#
# On the opposite hand, when you are equested for a chat, you just have to create the object but this time you must fill ip and port parameters. Then you can call in the same processus the connect() method which will connect to the user socket and return the socket id.
# From that point you can call process() method to get data from the opposite user and say() method to send him something.
#
# You can finally close the chat by calling abort() method
#
class DCCChat:
    def __init__(self,nick,ip="",port=-1):
        """Define a DCC chat connection
        The DCC chat represents a way to communicate with someone else through direct IP connection. It avoids talking through the server

        DCCChat object takes four arguments, the nickname of the user who has asked for a chat, the socket """
        self.nick = nick
        self.ip = ip
        self.port = int(port)
        self.sock = None
        self.buffer = ""
        self.queue = Queue()
        self_is_connected = 0
        
    def connect(self):
        if self.ip != "":
            try:
                self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
                self.sock.connect((self.ip,self.port))
                self.sock.setblocking(0)
                self_is_connected = 1
            except socket.error, sid:
                raise IRCConnectionError(str("Couldn't connect to socket " + self.ip + " on port " + str(self.port) + " : " + str(sid)))

        return self.sock.fileno()

    def get_sock_id(self):
        """Returns the socket id given by fileno()"""
        if self.sock != None:
            return self.sock.fileno()
        return -1
    
    def is_alive(self):
        """Return true if the connection has been established and is still alive"""
        return self_is_connected
    
    def say(self,msg):
        """Send a message through the socket. If the message doesn't end with a newline, it will be added"""
        if msg[-1] != '\n':
            msg = msg + "\n"
        self.sock.send(msg)

    def process(self):
        """Read incoming data on the socket"""
        try:
            self.buffer = self.sock.recv(512)
            self.queue.put(("read",self.nick,self.buffer))
        except socket.error, sid:
            self.is_connected = 0
            self.queue.put(("closed",self.nick))

    def has_more_messages(self):
        return not self.queue.empty()
   
    def get_next_message(self):
        return self.queue.get()
 
    def get_nick(self):
        """Return the nick of the opposite client"""
        return self.nick

    def get_socket(self):
        """Return the socket object itself or None if the connection hasn't been established yet"""
        return self.sock

    def set_socket(self,sock):
        """Set the socket for this connection"""
        self.sock = sock
        self_is_connected = 1
        
    def abort(self):
        """Close a chat connection"""
        self.sock.close()

#
# DCCSend gives the user the way to send files over a socket directly to an another user.
#
# You must in a first time create a DCCSend object by passing to it the user which will get the file and the filename. Then you should call the send() method in a separate thread and let it go until the all file has been sent.
#
#
class DCCSend:
    def __init__(self,to_whom,filename):
        """ """
        self.request_to_send_connexion = None
        self.running = 0
        self.to_whom = to_whom
        self.filename = filename

    #
    # The method will first create a socket and listen for a connection from the other user. The method will then send a private message to the opposite user to let him know that a DCC send request has been posted. If he accepts the DCC send, the connection will be established between the two users, and thus the file transfer will be proceed.
    #
    def send(self,connexion):
        """Send the file to the opposite client. This method will also send a PRIVMSG to the client before sending any data and will wait till the client accepts the connection"""
        filesize = os.path.getsize(self.filename)

        self.request_to_send_connexion = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.request_to_send_connexion.bind(('', 0))
        self.request_to_send_connexion.listen(1)

        port = self.request_to_send_connexion.getsockname()[1]
        addr = socket.gethostbyname(socket.gethostname()).split('.')
        ip = (int(addr[0]) << 24) | (int(addr[1]) << 16) | \
             (int(addr[2]) <<  8) | (int(addr[3]) <<  0)

        connexion.privmsg(self.to_whom,str(_x01 + "DCC SEND " + os.path.basename(self.filename) + " " + str(ip) + " " + str(port) + " " + str(filesize) + _x01) + "\r\n")

        try:
            acceptor, acceptor_host = self.request_to_send_connexion.accept()
            self.request_to_send_connexion.close()
            self.running = 1
            
            fileobject = file(self.filename,'rb')
            bytes_sent = 0
            amount = 0
            
            while self.running and bytes_sent < filesize:
                amount = acceptor.send(fileobject.read(DCC_PACKET_SIZE))
                data = acceptor.recv(4)
                bytes_sent = socket.ntohl(unpack("i", data)[0])
                    
            fileobject.close()
            acceptor.close()
        except:
            IRCConnectionError(str("Cant accept connection from " + to_whom))

    def abort(self):
        """Abort a file transfer"""
        if self.running:
            self.running = 0
        else:
            self.request_to_send_connexion.close()

#
# DCCGet gives the user the way to get a file from another user.
#
class DCCGet:
    def __init__(self,from_who,filename, ip, port, filesize):
        self.running = 0
        self.from_who = from_who
        self.filename= filename
        self.ip = ip
        self.port = port
        self.filesize = filesize

    # 
    def get(self):
        """Get a file which has been offered. It will connect first to the opposite client."""
        sender = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sender.connect((self.ip, int(self.port)))
        self.running = 1
        
        bytes_received = 0

        fileobject = file(self.filename,'wb')
        
        while self.running:
            buffer = sender.recv(DCC_PACKET_SIZE)
            if not buffer:
                break
            bytes_received = bytes_received + len(buffer)
            fileobject.write(buffer)
            fileobject.flush()
            sender.send(pack('l',socket.htonl(bytes_received)))
            
        fileobject.close()
        sender.close()
        
    def abort(self):
        """Abort a file transfer"""
        self.running = 0

class IRCServerConnexion:
    def __init__(self,server,port,nick,user,real_name,passwd='',host='localhost'):
        self.server = str(server)
        self.port = int(port)
        self.nickname = str(nick)
        self.username = str(user)
        self.host = str(host)
        self.real_name = str(real_name)
        self.passwd = str(passwd)
        self.connexion = None
        self.queue = Queue()
        self.channels = []
        self.is_connected = 0
        self.buffer = ""
        self.dcc_files_thread = {}

    #
    # This method will return either the socked id on success or will raise a IRCConnectionError
    #
    def connect(self):
        """Connect to the server using AF_INET and SOCK_STREAM. Once the connection has been done, it will set the server connection as alive. The socket is set in a non blocking mode. This methos will also send the password, the nick and the user message in this order to connect to the distant server. If one of these fails, it raises an IRCConnectionError exception. It will return the id of the socket which could be used with select.select() for instance."""
        self.connexion = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        try:
            self.connexion.connect((self.server,self.port))
            self.connexion.setblocking(0)
            if self.passwd != '':
                self.password(self.passwd)
            self.nick(self.nickname)
            self.user(self.username,self.host,self.server,self.real_name)
            self.is_connected = 1
        except socket.error, sid:
            raise IRCConnectionError(str("Couldn't connect to socket " + self.server + " on port " + str(self.port) + " : " + str(sid)))

        return self.connexion.fileno()
    
    def get_server(self):
        """Return the server name. It may change from the one passed to the constructor if the server redirects the connection on an another server."""
        return self.server

    def set_server(self,s):
        """Set the server name"""
        self.server = s

    def get_sock_id(self):
        return self.connexion.fileno()
    
    def get_port(self):
        """Return the port passed to the constructor"""
        return self.port

    def get_nick(self):
        """Return the nickname which will change each time users change their nick inside the client"""
        return self.nickname

    def set_nick(self,new_nick):
        self.nickname = new_nick
        
    def get_user(self):
        """Return the username"""
        return self.username

    def get_host(self):
        """Return the hostname"""
        return self.host

    def get_real_name(self):
        """Return the real name"""
        return self.real_name

    def get_password(self):
        """Return the password"""
        return self.passwd

    def get_channels(self):
        """Return the channels list which will be modified each time a user leaves or joins a channel"""
        return self.channels

    def get_socket(self):
        """Return the socket object or None if the connection is not over"""
        return self.connexion
    
    def __send_data(self,string):
        """Send data through the socket"""
        self.connexion.send(string + "\r\n")
    
    def is_alive(self):
        """Return true if the connection has been well established"""
        return self.is_connected

    def has_more_messages(self):
        """Return true if there are some messages queuing before being managed"""
        return not self.queue.empty()
   
    def get_next_message(self):
        """Return the next message on the queue. The message will be removed from it once read. WARNING : You must not use at the same time the normal and raw managing messages methods. It would screw up your application."""
        return self.queue.get()
 
    def has_more_raw_messages(self):
        """Return true if there are some messages queuing before being managed"""
        return not self.queue.empty()
   
    def get_next_raw_message(self):
        """Return the next message on the queue. The message will be removed from it once read"""
        return self.queue.get()
 
    def __get_data(self, bufsz = 512):
        """Return date read through the socket or '' if nothing appears on it"""
        try:
            buffer = self.connexion.recv(bufsz)
        except socket.error, sid:
            buffer = ""
        return buffer
    
    def __split_message(self, buffer):
        """Split data read on the socket in order to parse them correctly figuring out the different parts of the message. This method will do its best but many servers don't respect clearly the RFCs and thus are not well parsed. This is not a problem this module will deal with. It will return the different parts of the message as a tuple in this order :
        - the server name
        - the IRC command
        - the nickname from whom it comes if given or ''
        - the channel from where it comes if given or ''
        - the arguments passed to the command as they were given (which mean they will be parsed later)
        - the buffer wich ay contain some trailing data not parsed yet"""
        lines = _message_sep.split(buffer)
        buffer = lines[-1]
        lines = lines[:-1]

        #print lines
        
        prefix = []
        command = []
        arguments = []
        from_channel = []
        from_nickname = []
        
        for line in lines:
            m = _rfc_1459_command_regexp.match(line)
            if m.group("prefix"):
                s = string.split(m.group("prefix"), "!",1)
                if len(s) == 2:
                    from_nickname.append(s[0])
                    prefix.append(s[1])
                else:
                    from_nickname.append('')
                    prefix.append(s[0])
            else:
                from_nickname.append('')
                prefix.append(self.get_server())
                
            if m.group("command"):
                command.append(m.group("command"))
            else:
                command.append('')

            if m.group("argument"):
                a = string.split(m.group("argument"), " :", 1)
                from_channel.append(string.lstrip(a[0]))
                if len(a) == 2:
                    arguments.append(a[1] + '\r\n')
                else:
                   arguments.append('') 
            else:
                from_channel.append('')
                arguments.append('')
    
        return (prefix,command,from_nickname,from_channel,arguments,buffer) 

    def __decode_buffer(self, prefix,command,from_nickname,from_channel,arguments, buffer):
        """This is the core of the module. This method will read a splitted message from the socket and will put in the messages queue the result of the parsing"""
        i = 0
        while i < len(command):
            #print (prefix[i],command[i],from_nickname[i],from_channel[i],arguments[i]) 
            if command[i] == "PING":
                self.pong(self.get_server())
            elif command[i] == "PRIVMSG":
                if arguments[i][0] != _x01:
                    self.queue.put(("privmsg",self.get_server(),prefix[i],from_nickname[i],from_channel[i],arguments[i]))
                else:
                    ctcp_cmd = string.split(arguments[i],_x01)
                    if string.find(ctcp_cmd[1],"DCC SEND",0) == 0:
                        dcc_cmd = string.split(ctcp_cmd[1],' ')
                        iplong = long(dcc_cmd[3])
                        ip = str((iplong >> 24) & 0xFF) + '.' + str((iplong >> 16) & 0xFF) + '.' + \
                             str((iplong >> 8)  & 0xFF) + '.' + str((iplong >>  0) & 0xFF)
                        self.queue.put(("dccsend",self.get_server(),prefix[i],from_nickname[i],dcc_cmd[2],ip,dcc_cmd[4],dcc_cmd[5]))
                    elif string.find(ctcp_cmd[1],"DCC CHAT",0) == 0:
                        dcc_cmd = string.split(ctcp_cmd[1],' ')
                        iplong = long(dcc_cmd[3])
                        ip = str((iplong >> 24) & 0xFF) + '.' + str((iplong >> 16) & 0xFF) + '.' + \
                             str((iplong >> 8)  & 0xFF) + '.' + str((iplong >>  0) & 0xFF)
                        self.queue.put(("dccchat",self.get_server(),prefix[i],from_nickname[i],ip,dcc_cmd[4]))
                    else:
                        data = string.split(ctcp_cmd[1],' ',1)
                        if len(data) == 1:
                            self.queue.put(("ctcp",self.get_server(),prefix[i],from_nickname[i],from_channel[i],data[0],''))
                        else:
                            self.queue.put(("ctcp",self.get_server(),prefix[i],from_nickname[i],from_channel[i],data[0],data[1]))
            elif command[i] == "NOTICE":
                self.queue.put(("notice",self.get_server(),prefix[i],from_nickname[i],from_channel[i],arguments[i]))
            elif command[i] == "JOIN":
                self.queue.put(("join",self.get_server(),prefix[i],from_nickname[i],arguments[i]))
            elif command[i] == "KICK":
                from_ch = string.split(from_channel[i],' ',1)
                self.queue.put(("kick",self.get_server(),prefix[i],from_nickname[i],from_ch[0],arguments[i]))
            elif command[i] == "PING":
                self.queue.put(("ping",arguments[i]))
            elif command[i] in ["TOPIC","332"]:
                channel = from_channel[i]
                who = from_nickname[i]
                if from_channel[i][0] not in ['#','&','!']:
                    pos = string.find(from_channel[i],'#')
                    channel = from_channel[i][pos:]
                    who = from_channel[i][:pos]
                self.queue.put(("topic",self.get_server(),prefix[i],who,channel,arguments[i]))
            elif command[i] == "331":
                data = string.split(from_channel[i],' ',1)
                self.queue.put(("notopic",self.get_server(),prefix[i],data[0],data[1],arguments[i]))
            elif command[i] == "NICK":
                self.queue.put(("nick",self.get_server(),prefix[i],from_nickname[i],arguments[i]))
            elif command[i] in ["353","366"]:
                if command[i] == "353":
                    channel = from_channel[i]
                    pos = string.find(from_channel[i],'#')
                    if pos != -1:
                        channel = from_channel[i][pos:]
                    self.queue.put(("datanames",prefix[i],channel,string.split(arguments[i][:-3], ' ')))
                else:
                    self.queue.put(("endnames",prefix[i],arguments[i]))
            elif command[i] == "372":
                self.queue.put(("datamotd",prefix[i],arguments[i]))
            elif command[i] == "375":
                self.queue.put(("startmotd",prefix[i],arguments[i]))
            elif command[i] == "376":
                self.queue.put(("endmotd",prefix[i],arguments[i]))
            elif command[i] in ["001","002","003"]:
                self.queue.put(("init_info",prefix[i],command[i],arguments[i]))
            elif command[i] == "004":
                responses_004 = string.split(from_channel[i],' ')
                if len(responses_004) == 5:
                    responses_004.append('')
                self.queue.put(("init_info",prefix[i],command[i],responses_004[1],responses_004[2],responses_004[3],responses_004[4],responses_004[5]))
            elif command[i] == "005":
                response_005 = string.split(from_channel[i],' ',1)
                responses_005 = string.split(response_005[1], ' ')
                self.queue.put(("init_info",prefix[i],command[i],responses_005))
            elif command[i] in ["251","252","253","254","255"]:
                if command[i] in ["251","255"]:
                    self.queue.put(("luser",prefix[i],command[i],arguments[i]))
                else:
                    value = string.split(from_channel[i],' ')
                    self.queue.put(("luser",prefix[i],command[i],value[1] + " " + arguments[i]))
            elif command[i] == "303":
                if arguments != '\r\n':
                    self.queue.put(("ison",prefix[i],arguments[i][:-3]))
                else:
                    self.queue.put(("ison",prefix[i],''))
            elif command[i] in ["MODE","221"]:
                if prefix[i] == self.get_nick():
                    self.queue.put(("mode",self.get_server(),from_channel[i],arguments[i]))
                else:
                    mode_params = string.split(from_channel[i],' ')
                    if len(mode_params) == 2:
                        mode_params.append('')
                    self.queue.put(("mode",self.get_server(),prefix[i],from_nickname[i],mode_params[0],mode_params[1],mode_params[2]))
            elif command[i] in ["321","322","323"]:
                if command[i] == "322":
                    chann_data = string.split(from_channel[i], ' ')
                    if arguments[i] != '\r\n':
                        comment = arguments[i]
                    else:
                        comment = ''
                    self.queue.put(("datalist",prefix[i],chann_data[1],chann_data[2],comment))
                elif command[i] == "321":
                    self.queue.put(("startlist",prefix[i],arguments[i]))
                elif command[i] == "323":
                    self.queue.put(("endlist",prefix[i],arguments[i]))
            elif command[i] == "311":
                data = string.split(from_channel[i],' ')
                self.queue.put(("datawhois",prefix[i],command[i],data[1],data[2],data[3],data[4],arguments[i]))
            elif command[i] in ["312","317"]:
                data = string.split(from_channel[i],' ')
                self.queue.put(("datawhois",prefix[i],command[i],data[1],data[2],arguments[i]))
            elif command[i] in ["313","318","319"]:
                print (prefix[i],command[i],from_nickname[i],from_channel[i],arguments[i]) 
                data = string.split(from_channel[i],' ')
                self.queue.put(("datawhois",prefix[i],command[i],data[1],arguments[i]))
            elif command[i] == "318":
                data = string.split(from_channel[i],' ')
                self.queue.put(("endwhois",prefix[i],data[1],arguments[i]))
            elif command[i] == "314":
                data = string.split(from_channel[i],' ')
                self.queue.put(("datawhowas",prefix[i],data[1],data[2],data[3],arguments[i]))
            elif command[i] == "369":
                self.queue.put(("endwhowas",prefix[i],arguments[i]))
            elif command[i] == "351":
                data = string.split(from_channel[i],' ')
                if len(data) == 3:
                    self.queue.put(("version",prefix[i],data[0],data[1],data[2],''))
                else:
                    self.queue.put(("version",prefix[i],data[0],data[1],data[2],data[3]))                    
            elif command[i] == "352":
                data = string.split(from_channel[i],' ',1)
                data = string.split(data[1],' ')
                extra_data = string.split(arguments[i],' ')
                self.queue.put(("datawho",prefix[i],data[0],data[1],data[2],data[3],data[4],data[5],extra_data[0],extra_data[1]))
            elif command[i] == "315":
                data = string.split(from_channel[i],' ',1)
                self.queue.put(("endwho",prefix[i],data[1],arguments[i]))
            elif command[i] == "INVITE":
                self.queue.put(("invite",self.get_server(),prefix[i],from_nickname[i],from_channel[i],arguments[i]))
            elif command[i] == "PART":
                self.queue.put(("part",self.get_server(),prefix[i],from_nickname[i],from_channel[i],arguments[i]))
            elif command[i] == "QUIT":
                self.queue.put(("quit",self.get_server(),prefix[i],from_nickname[i],arguments[i]))
            elif command[i] in ["211","212","213","214","215","216","218","241","242","243","244"]:
                pass
            elif command[i] in ["200","201","202","203","204","205","206","208","261"]:
                self.queue.put(("datatrace",self.get_server(),command[i],from_channel[i]))
            elif command[i] == "262":
                self.queue.put(("endtrace",self.get_server(),from_channel[i],arguments[i]))
            elif command[i] == "364":
                mask_server = string.split(from_channel[i],' ')
                dst_counter_info = string.split(arguments[i],' ')
                self.queue.put(("datalinks",prefix[i],mask_server[1],mask_server[2],dst_counter_info[0],dst_counter_info[1]))
            elif command[i] == "365":
                mask_server = string.split(from_channel[i],' ')
                self.queue.put(("endlinks",prefix[i],mask_server[1],arguments[i]))
            elif command[i] == "391":
                asked_server = string.split(from_channel[i],' ')
                self.queue.put(("time",prefix[i],asked_server[1],arguments[i]))
            elif command[i] == "306":
                self.queue.put(("markedaway",prefix[i],from_channel[i],arguments[i]))
            elif command[i] == "305":
                self.queue.put(("unmarkedaway",prefix[i],from_channel[i],arguments[i]))
            elif command[i] == "301":
                who = string.split(from_channel[i],' ')
                self.queue.put(("isaway",prefix[i],who[0],who[1],arguments[i]))
            elif command[i] == "263":
                self.queue.put(("tryagain",prefix[i],from_channel[i],arguments[i]))
            elif command[i] == "302":
                users_data = string.split(arguments[i][:-3],' ')
                for user in users_data:
                    self.queue.put(("userhost",prefix[i],from_channel[i],user))
            elif command[i] == "371":
                self.queue.put(("datainfo",prefix[i],from_channel[i],arguments[i]))
            elif command[i] == "374":
                self.queue.put(("endinfo",prefix[i],from_channel[i],arguments[i]))
            elif command[i] in ["256","257","258","259"]:
                self.queue.put(("admin",prefix[i],from_channel[i],arguments[i]))
            elif command[i] == "333":
                your_nick,channel,creator,timestamp = string.split(from_channel[i],' ')
                self.queue.put(("topic_info",prefix[i],your_nick,channel,creator,timestamp))
            elif command[i] in _error_range:
                request = string.split(from_channel[i],' ')
                if len(request) >= 2:
                    self.queue.put(("error",prefix[i],command[i],request[1],arguments[i]))
                else:
                    self.queue.put(("error",prefix[i],command[i],'',arguments[i]))
            else:
               self.queue.put(("unknown",prefix[i],command[i],from_nickname[i],from_channel[i],arguments[i]))
            
            i = i + 1
            
    def process_raw(self):
        """Public method to call to get raw incoming data"""
        if self.is_connected:
            buffer = self.__get_data()
            if buffer != "":
                self.queue.put(buffer)
                
    def process(self):
        """Public method to call to process incoming data on the socket"""
        if self.is_connected:
            buffer = self.__get_data()
            if buffer != "":
                self.buffer = self.buffer + buffer
                prefix,command,from_nickname,from_channel,arguments,self.buffer = self.__split_message(self.buffer)
                self.__decode_buffer(prefix,command,from_nickname,from_channel,arguments,self.buffer)
                
    def nick(self, nickname):
        """NICK message is used to give user a nickname or change the previous one."""
        self.__send_data("NICK " + nickname)
        
    def user(self, username, hostname, server, real_name):
        """Shouldn't be explictely used. The USER message is used at the beginning of connection to specif the username, hostname, servername and realname of s new user.  It is also used in communication between servers to indicate new user arriving on IRC, since only after both USER and NICK have been received from a client does a user become registered."""
        self.__send_data("USER %s %s %s :%s" % (username, hostname, server, real_name))
        
    def disconnect(self):
        """Disconnect from a server. Will close the socket"""
        if self.is_connected:
            self.connexion.close()
            self.connexion = None
                
    def quit(self, message=""):
        """A client session is ended with a quit message.  The server must close the connection to a client which sends a QUIT message. If a message is given, this will be sent instead of the default message,the nickname."""
        self.is_connected = 0
        command = ""
        if message == "":
            command = "QUIT :" + self.get_nick()
        else:
            command = "QUIT :" + message
        self.__send_data(command)

    def oper(self, passwd):
        """OPER message is used by a normal user to obtain operator privileges. This method will use the nick provided by get_nick()"""
        self.__send_data("OPER %s %s" % (self.get_nick(),passwd))

    def password(self, passwd):
        """The PASS command is used to set a 'connection password'.  The password can and must be set before any attempt to register the connection is made."""
        self.__send_data("PASS " + passwd)

    def mode_user(self, user, modes):
        """The user MODEs are typically changes which affect either how the client is seen by others or what 'extra' messages the client is sent.See the RFC 1459 for further details"""
        self.__send_data("MODE %s %s" % (user,string.join(modes, "")))

    def mode_channel(self, channel, modes, user="", limit="", mask="", key=""):
        """The MODE command is provided so that channel operators may change the characteristics of `their' channel. See the RFC 1459 for further details"""
        if channel[0] == '+':
            return
        mode_command = "MODE " + channel + " " + string.join(modes, "")
        _modes = string.split(modes[1:])
        
        if "p" in _modes or "s" in _modes or "i" in _modes or "t" in _modes or "n" in _modes or "m" in _modes:
            self.__send_data(mode_command)
        elif "o" in _modes or "v" in _modes:
            if user != "":
                mode_command = mode_command + " " + user
                self.__send_data(mode_command)
        elif "l" in _modes:    
            if limit != "":
                mode_command = mode_command + " " + limit
                self.__send_data(mode_command)
        elif "b" in _modes:
            if mask != "":
                mode_command = mode_command + " " + mask
                self.__send_data(mode_command)
        elif "k" in _modes:
            if key != "":
                mode_command = mode_command + " " + key
                self.__send_data(mode_command)

    def version(self, server=""):
        """The VERSION message is used  to  query  the  version  of  the  server program."""
        if server == "":
            server = self.get_server()
        self.__send_data("VERSION " + server)

    def stats(self, request, server=""):
        """The stats message is used to query statistics of certain server. Request can be one of these (see RFC 1459):
        c - returns a list of servers which the server may connect to or allow connections from;
        h - returns a list of servers which are either forced to be treated as leaves or allowed to act as hubs;
        i - returns a list of hosts which the server allows a client to connect from;
        k - returns a list of banned username/hostname combinations for that server;
        l - returns a list of the server's connections, showing how long each connection has been established and the traffic over that connection in bytes and messages for each direction;
        m - returns a list of commands supported by the server and the usage count for each if the usage count is non zero;
        o - returns a list of hosts from which normal clients may become operators;
        y - show Y (Class) lines from server's configuration file;
        u - returns a string showing how long the server has been up."""
        stats_command = "STATS " + request
        if server != "":
            stats=command = stats_command + " " + server
        self.__send_data(stats_command)
        
    def join(self, targets, keys=None):
        """The JOIN command is used by client to start listening a specific channel.The conditions which affect this are as follows:
           1.  the user must be invited if the channel is invite-only;
           2.  the user's nick/username/hostname must not match any active bans;
           3.  the correct key (password) must be given if it is set."""
        if len(targets) == 0:
            self.__send_data("JOIN 0")
        else:
            join_command = "JOIN "
            if type(targets) is StringType:
                self.channels.append(targets)
                if keys:
                    if type(keys) is StringType:
                        join_command = join_command + targets + " " + keys
                    elif type(keys) is ListType:
                        join_command = join_command + targets + " " + keys[0]
                else:
                    join_command = join_command + targets
            elif type(targets) is ListType:
                self.channels.extend(targets)
                if keys:
                    keys_string = ""
                    if type(keys) is StringType:
                        keys_string = keys
                    elif type(keys) is ListType:   
                        keys_string = string.join(keys, ",")
                    targets_string = string.join(targets, ",")
                    join_command = join_command + targets_string + " " + keys_string
                else:
                    targets_string = string.join(targets, ",")
                    join_command = join_command + targets_string
            self.__send_data(join_command)
                    
    def privmsg(self, targets, message):
        """Send a private message to some targets (channels or users given as a string or a list)"""
        if type(targets) is StringType:
            self.__send_data("PRIVMSG %s :%s" % (targets, message))
        elif type(targets) is ListType:
            self.__send_data("PRIVMSG %s :%s" % (string.join(targets, ","), message))

    def notice(self, targets, message):
        """Send a notice to some targets (channels or users given as a string or a list)"""
        if type(targets) is StringType:
            self.__send_data("NOTICE %s :%s" % (targets, message))
        elif type(targets) is ListType:
            self.__send_data("NOTICE %s :%s" % (string.join(targets, ","), message))

    def invite(self, username, channel):
        """The INVITE message is used to invite users to a channel."""
        self.__send_data("INVITE " + username + " " + channel)

    def names(self, targets):
        """By using the NAMES command, a user can list all nicknames that are visible to them on any channel that they can see. (Targets can be a string or a list)"""
        if type(targets) is StringType:
            self.__send_data("NAMES " + targets)
        elif type(targets) is ListType:
            self.__send_data("NAMES %s" % string.join(targets, ","))   

    def list(self, targets = ""):
        """The list message is used to list channels and their topics. (Targets can be a string or a list)"""
        if type(targets) is StringType:
            if targets == "":
                self.__send_data("LIST")
            else:
                self.__send_data("LIST " + targets)
        elif type(targets) is ListType:
            self.__send_data("LIST %s" % string.join(targets,","))

    def kick(self, channel, user, comment = ""):
        """The KICK command can be  used  to  forcibly  remove  a  user  from  a channel."""
        kick_command = "KICK " + channel + " " + user
        if comment != "":
            kick_command = kick_command + " :" + comment
        self.__send_data(kick_command)

    def topic(self, target, topic=""):
        """The TOPIC message is used to change or view the topic of a channel."""
        if topic == "":
            self.__send_data("TOPIC " + target)
        else:
            self.__send_data("TOPIC %s :%s" % (target, topic))

    def part(self, targets,message=""):
        """The PART message causes the client sending the message to be removed from the list of active users for all given channels listed in the parameter string"""
        if type(targets) is StringType:
            if message == "":
                self.__send_data("PART " + targets)
            else:
                self.__send_data("PART %s %s" % (targets,message))
        elif type(targets) is ListType:
            if message == "":
                self.__send_data("PART %s" % string.join(targets, ",")) 
            else:
                self.__send_data("PART %s %s" %  (string.join(targets, ","),message)) 

    def links(self, server="", mask=""):
        """With LINKS, a user can list all servers which are known by the server answering the query."""
        links_command = "LINKS"
        if server != "":
            links_command = links_command + " " + server
        if mask != "":
            links_command = links_command + " " + mask
        self.__send_data(links_command)

    def time(self, server=""):
        """The time message is used to query local time from the specified server."""
        if server == "":
            server = self.get_server()
        self.__send_data("TIME " + server)

    def userhost(self, targets):
        """The USERHOST command takes a list of up to 5 nicknames, each separated by a space character and returns a list of information about each nickname that it found. """
        userhost_command = "USERHOST "
        if type(targets) is StringType:
            userhost_command = userhost_command + targets
        elif type(targets) is ListType:
            userhost_command = userhost_command + string.join(targets, " ")
        self.__send_data(userhost_command)
        
    def trace(self, target=""):
        """TRACE command is used to find the route to specific server."""
        if target == "":
            target = self.get_server()
        self.__send_data("TRACE " + target)

    def admin(self, server=""):
        """The admin message is used to find the name of the administrator of the given server, or current server if <server> parameter is omitted."""
        if server == "":
            server = self.get_server()
        self.__send_data("ADMIN " + server)

    def info(self, target=""):
        """The INFO command is required to return information which describes the server: its version, when it was compiled, the patchlevel, when it was started, and any other miscellaneous information which may be considered to be relevant. """
        if target == "":
            target = self.get_server()
        self.__send_data("INFO " + target)

    def who(self, targets, only_ops=0):
        """The WHO message is used by a client to generate a query which returns a list of information which 'matches' the targets parameter given by the client. If the only_ops parameter is set to 1 only operators are returned according to the targets mask supplied."""
        who_command = "WHO " + targets
        if only_ops != 0:
            who_command = who_command + " o"
        self.__send_data(who_command)

    def whois(self, targets, server=""):
        """This message is used to query information about particular user.""" 
        whois_command = "WHOIS "
        if server != "":
            whois_command = whois_command + server + " "
        if type(targets) is StringType:
            whois_command = whois_command + targets
        elif type(targets) is ListType:
            whois_command = whois_command + string.join(targets, ",")
        self.__send_data(whois_command)

    def whowas(self, target, max=1, server=""):
        """Whowas asks for information about a nickname which no longer exists."""
        whowas_command = "WHOWAS " + target + " " + str(max)
        if server != "":
            whowas_command = whowas_command + " " + server
        self.__send_data(whowas_command)

    def kill(self, target, comment=""):
        """The KILL message is used to cause a client-server connection to be closed by the server which has the actual connection."""
        kill_command = "KILL " + target
        if comment != "":
            kill_command = kill_command + " " + comment
        self.__send_data(kill_command)

    def ping(self, targets=None):
        """The PING message is used to test the presence of an active client at the other end of the connection."""
        if targets == None:
            targets = self.get_server()
        if type(targets) is StringType:
            self.__send_data("PING " + targets)
        elif type(targets) is ListType:
            self.__send_data("PING %s %s" % string.join(targets, " "))

    def pong(self, from_daemon, to_daemon=""):
        """PONG message is a reply to ping message. (Implemented as set in RFC 1459"""
        pong_command = "PONG " + from_daemon
        if to_daemon != "":
            pong_command = pong_command + to_daemon
        self.__send_data(pong_command)

    def raw(self, data):
        """Send a raw command to the server"""
        self.__send_data(data)

    def away(self, message=""):
        """With the AWAY message, clients can set an automatic reply string for any PRIVMSG commands directed at them (not to a channel they are on)."""
        away_command = "AWAY"
        if message != "":
            away_command = away_command + " " + message
        self.__send_data(away_command)

    def rehash(self):
        """The rehash message can be used by the operator to force the server to re-read and process its configuration file. """
        self.__send_data("REHASH")

    def restart(self):
        """ """
        self.__send_data("RESTART")

    def ison(self, targets):
        """The ISON command gets a response about whether a given nickname was currently on IRC."""
        ison_command = "ISON "
        if type(targets) is StringType:
            ison_command = ison_command + targets
        elif type(targets) is ListType:
            ison_command = ison_command + string.join(targets,' ')
        self.__send_data(ison_command)

    def ctcp_version_request(self,nick):
        """Send a CTCP version request to a user"""
        self.privmsg(nick,str(_x01 + "VERSION" + _x01))

    def ctcp_version(self,to_whom,client_name,client_version,client_arch):
        """Reply to a CTCP version request. You must provide info about your client and other stuff"""
        self.notice(to_whom,str(_x01 + "VERSION " + client_name + ":" + client_version + ":" + client_arch + _x01))

    def ctcp_finger(self):
        """ """
        pass

    def ctcp_finger_request(self,to_whom):
        """ """
        pass
    
    def ctcp_source(self):
        """ """
        pass

    def ctcp_source_request(self,to_whom):
        """ """
        pass

    def ctcp_action(self,target,msg):
        """ """
        self.privmsg(target,str(_x01 + "ACTION " + msg + _x01))

    def ctcp_error(self, to_whom,command, errormessage):
        """ """
        self.notice(to_whom,str(_x01 + "ERRMSG " + command + " :" + errormessage + _x01))

    def ctcp_ping(self, to_whom,timestamp):
        """Reply to a CTCP ping request. You must provide the timestamp given in the CTCP time request"""
        self.notice(to_whom,str(_x01 + "PING " + str(timestamp) + _x01))

    def ctcp_ping_request(self, to_whom,timestamp):
        """Request a CTCP ping reply. You must provide the timestamp given in the CTCP time request"""
        self.privmsg(to_whom,str(_x01 + "PING " + str(timestamp) + _x01))

    def ctcp_time(self, to_whom,rightnow):
        """Reply to a CTCP time request"""
        self.notice(to_whom,str(_x01 + "TIME " + rightnow + _x01))
        
    def ctcp_time_request(self, to_whom,rightnow):
        """Request a CTCP time reply"""
        self.privmsg(to_whom,str(_x01 + "TIME " + rightnow + _x01))

    def ctcp_time_request(self, to_whom):
        """Request a CTCP time reply"""
        self.privmsg(to_whom,str(_x01 + "TIME" + _x01))
        
    def ctcp_user_info(self,to_whom,userinfo):
        """Reply to CTCP userinfo request"""
        self.notice(to_whom,str(_x01 + "USERINFO :" + userinfo + _x01))

    def ctcp_user_info_request(self,to_whom):
        """Request CTCP userinfo reply"""
        self.privmsg(to_whom,str(_x01 + "USERINFO" + _x01))
        
    def ctcp_client_info(self):
        """ """
        pass
    
    def ctcp_client_info_request(self,to_whom):
        """ """
        pass

    def dcc_chat_request(self,to_whom):
        """Request a DCC chat connection to an user"""
        request_to_send_connexion = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        request_to_send_connexion.bind(('', 0))
        request_to_send_connexion.listen(1)

        port = request_to_send_connexion.getsockname()[1]
        addr = socket.gethostbyname(socket.gethostname()).split('.')
        ip = (int(addr[0]) << 24) | (int(addr[1]) << 16) | (int(addr[2]) <<  8) | (int(addr[3]) <<  0)

        self.privmsg(to_whom, str(_x01 + "DCC CHAT chat " + str(ip) + " " + str(port) + _x01))
        
        try:
            acceptor, acceptor_host = request_to_send_connexion.accept()
            request_to_send_connexion.close()
            self.queue.put(("dccchatok",self.get_server(),to_whom,acceptor))
        except:
            raise IRCConnectionError(str("Can't accept connexion from " + to_whom))
            
#
# End of the pyirc module
#
